namespace Pic.DAL.SqlServer
{
    #region Using directives (this needs to be here to avoid error with 'MSLinqToSQLGenerator')
    using System;
    using System.Data;
    using System.Data.SqlClient;
    using System.Text;
    using System.Collections.Generic;

    using System.Linq;
    using System.Data.Linq;
    using System.Data.Linq.Mapping;

    using System.Drawing;
    using System.Drawing.Imaging;
    #endregion

    #region ExceptionDAL
    public class ExceptionDAL : System.Exception
    {
        public ExceptionDAL() { }
        public ExceptionDAL(string message) : base(message) { }
        public ExceptionDAL(string message, System.Exception innerException) : base(message, innerException) { }
    }
    #endregion

    #region PPDataContext
    partial class PPDataContext
    {
        public static void RemoveRecord<T>(T recordToRemove) where T : class
        {
            // retrieve PPDataContext instance
            PPDataContext db = new PPDataContext();

            Table<T> tableData = db.GetTable<T>();
            var deleteRecord = tableData.SingleOrDefault(record => record == recordToRemove);
            if (deleteRecord != null)
            {
                tableData.DeleteOnSubmit(deleteRecord);
            }
            // submit changes now
            db.SubmitChanges();
        }
    }
    #endregion

    #region CardboardProfile
    public partial class CardboardProfile
    {
        public static CardboardProfile[] GetAll(PPDataContext db)
        {
            IEnumerable<CardboardProfile> profiles = from cp in db.CardboardProfiles select cp;
            return profiles.ToArray();
        }

        public static CardboardProfile CreateNew(PPDataContext db, string name, string code, float thickness)
        {
            // is already existing ?
            if (db.CardboardProfiles.Count(cp => cp.Name.ToLower() == name.ToLower()) > 0)
                throw new ExceptionDAL(string.Format("CardboardProfile with name = {0} already exists", name));
            if (db.CardboardProfiles.Count(cp => cp.Code.ToLower() == code.ToLower()) > 0)
                throw new ExceptionDAL(string.Format("CardboardProfile with code = {0} already exists", code));
            // build new cardboard profile
            CardboardProfile profile = new CardboardProfile();
            profile.Name = name;
            profile.Code = code;
            profile.Thickness = thickness;
            // insert
            db.CardboardProfiles.InsertOnSubmit(profile);
            db.SubmitChanges();
            // retrieve new cardboard profile
            return db.CardboardProfiles.Single(cp => cp.Code == code);
        }

        public bool HasMajorationSets
        {
            get { return MajorationSets.Count > 0; }
        }

        public static CardboardProfile GetByName(PPDataContext db, string name)
        {
            return db.CardboardProfiles.SingleOrDefault(cbp => cbp.Name.ToLower() == name.ToLower());
        }

        public void Delete(PPDataContext db)
        {
            // check that this profile is not used in majoration sets that the only 
            // check that no component has a single majoration set that depends on this
            // delete majoration sets
            foreach (MajorationSet majoSet in this.MajorationSets)
                majoSet.Delete(db);
            // delete profile itself
            db.CardboardProfiles.DeleteOnSubmit(this);
            db.SubmitChanges();
        }

        public static void DeleteAll(PPDataContext db)
        {
            foreach (CardboardProfile cbp in db.CardboardProfiles)
                cbp.Delete(db);
        }

        public void Update()
        {
            PPDataContext db = new PPDataContext();
            if (db.CardboardProfiles.Count(cp => cp.ID != _ID && cp.Name == _Name) > 0)
                throw new ExceptionDAL(string.Format("A cardboard profile with name {0} already exists", _Name));
            if (db.CardboardProfiles.Count(cp => cp.ID != _ID && cp.Code == _Code) > 0)
                throw new ExceptionDAL(string.Format("A cardboard profile with code {0} already exists", _Code));
            db.SubmitChanges();
        }

        public override string ToString()
        {
            return string.Format("Name={0} Code={1} Thickness={2}", _Name, _Code, _Thickness);
        }
    }
    #endregion

    #region Cardboard format
    public partial class CardboardFormat
    {
        public static CardboardFormat CreateNew(PPDataContext db, string name, string description, float length, float width)
        {
            // is already existing ?
            if (db.CardboardFormats.Count(cf => cf.Name.ToLower() == name.ToLower()) > 0)
                throw new ExceptionDAL(string.Format("CardboardFormat with name = {0} already exists", name));
            // build new cardboard format
            CardboardFormat cardboardFormat = new CardboardFormat();
            cardboardFormat.Name = name;
            cardboardFormat.Description = description;
            cardboardFormat.Length = length;
            cardboardFormat.Width = width;
            db.CardboardFormats.InsertOnSubmit(cardboardFormat);
            db.SubmitChanges();
            return cardboardFormat;
        }

        public static CardboardFormat[] GetAll(PPDataContext db)
        {
            return (from cbf in db.CardboardFormats select cbf).ToArray();
        }

        public static CardboardFormat GetByName(PPDataContext db, string name)
        {
            return db.CardboardFormats.Single(cbf => cbf.Name.ToLower() == name.ToLower());
        }

        public static CardboardFormat GetByID(PPDataContext db, int id)
        {
            return db.CardboardFormats.Single(cbf => cbf.ID == id);
        }

        public void Delete(PPDataContext db)
        {
            db.CardboardFormats.DeleteOnSubmit(this);
            db.SubmitChanges();
        }

        public static void DeleteAll(PPDataContext db)
        {
            db.CardboardFormats.DeleteAllOnSubmit(from cbf in db.CardboardFormats select cbf);
            db.SubmitChanges();
        }

        public override string ToString()
        {
            return string.Format("Name={0} Length={1} Width={2}", _Name, _Length, _Width);
        }
    }
    #endregion

    #region MajorationSet
    public partial class MajorationSet
    {
        Dictionary<string, double> Dictionnary
        {
            get
            {
                Dictionary<string, double> dict = new Dictionary<string, double>();
                foreach (Majoration majo in this.Majorations)
                    dict.Add(majo.Name, (double)majo.Value);
                return dict;
            }
        }

        public void Delete(PPDataContext db)
        {
            db.Majorations.DeleteAllOnSubmit(db.Majorations.Where(m => m.MajorationSetID == this.ID));
            db.MajorationSets.DeleteOnSubmit(db.MajorationSets.Single(mjs => mjs.ID == ID));
            db.SubmitChanges();
        }
    }
    #endregion

    #region DocumentType
    public partial class DocumentType
    {
        public static DocumentType CreateNew(PPDataContext db, string name, string description, string application)
        {
            // already exists ?
            if (db.DocumentTypes.Count(dtype => dtype.Name.ToLower() == name.ToLower()) > 0)
                throw new ExceptionDAL(string.Format("Document type with Name = {0} already exists", name));
            // create new document type
            DocumentType docType = new DocumentType();
            docType.Name = name;
            docType.Description = description;
            docType.Application = application;
            db.DocumentTypes.InsertOnSubmit(docType);
            db.SubmitChanges();
            // retrieve newly created document type and return
            return db.DocumentTypes.Single(dtype => dtype.Name == name);
        }

        public static DocumentType GetByName(PPDataContext db, string name)
        {
            // retrieve newly created document type 
            return db.DocumentTypes.Single(dtype => dtype.Name.ToLower() == name.ToLower());
        }

        public override string ToString()
        {
            return string.Format("Name={0} Description={1} Application={2}", _Name, _Description, _Application);
        }

        public void Delete(PPDataContext db)
        {
            // delete document of this type
            var docQuery =  from doc in db.Documents
                            where (doc.DocumentType.ID ==  this.ID)
                            select doc;
            foreach (Document doc in docQuery)
                doc.Delete(db);
            // delete document type itself
            db.DocumentTypes.DeleteOnSubmit(this);
            db.SubmitChanges();
        }

        public static void DeleteAll(PPDataContext db)
        {
            foreach (DocumentType docType in db.DocumentTypes)
                docType.Delete(db);
        }
    }
    #endregion

    #region File
    public partial class File
    {
        #region Create and delete
        public static File CreateNew(PPDataContext db, string filePath)
        {
            // check if file is actually existing
            if (!System.IO.File.Exists(filePath))
                throw new ExceptionDAL(string.Format("File {0} can not be found.", filePath));
            // get file extension
            string fileExt = System.IO.Path.GetExtension(filePath);
            // copy to repository
            System.Guid guid = System.Guid.NewGuid();
            string repPath = System.IO.Path.Combine(RepositoryPath, guid.ToString().Replace('-', '_'));
            repPath = System.IO.Path.ChangeExtension(repPath, fileExt);
            System.IO.File.Copy(filePath, repPath);
            // create File object
            File file = new File();
            file._Extension = fileExt;
            file._Guid = guid;
            file._DateCreated = System.DateTime.Now;
            // insert file
            db.Files.InsertOnSubmit(file);
            db.SubmitChanges();
            // retrieve newly created file
            return db.Files.Single(f => f.Guid == guid);
        }

        internal void Delete(PPDataContext db)
        {
            // delete actual file
            System.IO.File.Delete(Path);
            // delete record
            db.Files.DeleteOnSubmit(this);
            db.SubmitChanges();
        }
        #endregion

        #region Additionnal properties
        private static string RepositoryPath
        { get { return global::Pic.DAL.SqlServer.Properties.Settings.Default.Repository; } }

        public string Path
        {
            get
            {
                string filePath = System.IO.Path.Combine(File.RepositoryPath, _Guid.ToString().Replace('-', '_'));
                filePath = System.IO.Path.ChangeExtension(filePath, _Extension);
                return filePath;
            }
        }
        #endregion
    }
    #endregion

    #region Thumbnail
    public partial class Thumbnail
    {
        #region Create and delete
        public static Thumbnail CreateNew(PPDataContext db, string imageFilePath)
        {
            // build new thumbnail object
            Thumbnail thumbnail = new Thumbnail();
            thumbnail.File = File.CreateNew(db, imageFilePath);
            thumbnail.MimeType = GetMimeType(IMAGEFORMAT);
            thumbnail.Width = THUMBNAILSIZE;
            thumbnail.Height = THUMBNAILSIZE;
            thumbnail.ThumbCache = BmpToBytes(Image.FromFile(imageFilePath), ImageFormat.Bmp);
            // insertion
            db.Thumbnails.InsertOnSubmit(thumbnail);
            db.SubmitChanges();
            return thumbnail;
        }

        internal void Delete(PPDataContext db)
        {
            // delete thumbnail record
            db.Thumbnails.DeleteOnSubmit(this);
            db.SubmitChanges();
            // delete attached file
            File.Delete(db);
        }
        #endregion

        #region Image (Thumbnail) helpers
        /// <summary>
        /// get the thumbnail as stored in the database
        /// </summary>
        /// <returns></returns>
        public Image GetImage()
        {
            return Image.FromStream(new System.IO.MemoryStream(ThumbCache.ToArray()));
        }

        public static void Annotate(Image image, string annotation)
        {
            // graphics
            Graphics grph = Graphics.FromImage(image);
            Font tfont = new Font("Arial", _fontSize);
            Brush tbrush = new SolidBrush(_fontColor);
            StringFormat sf = new StringFormat();
            sf.Alignment = StringAlignment.Far;
            sf.LineAlignment = StringAlignment.Far;
            Size txtSize = grph.MeasureString(annotation, tfont).ToSize();
            grph.FillRectangle(new SolidBrush(_backgroundColor), new Rectangle(image.Width - txtSize.Width - 2, image.Height - txtSize.Height - 2, txtSize.Width + 2, txtSize.Height + 2));
            grph.DrawString(annotation, tfont, tbrush, new PointF(image.Width - 2, image.Height - 2), sf);
        }

        /// <summary>
        /// returns a byte[] array which represents the input Image
        /// </summary>
        /// <param name="bmp">The source image to return as byte[] array</param>
        /// <returns>byte[] array which represents the input Image</returns>
        private static System.Data.Linq.Binary BmpToBytes(Image img, ImageFormat imageFormat)
        {
            System.IO.MemoryStream ms = new System.IO.MemoryStream();
            // Save to memory using the Jpeg format
            img.Save(ms, imageFormat);
            // read to end
            byte[] bmpBytes = ms.GetBuffer();
            return new System.Data.Linq.Binary(bmpBytes);
        }

        /// <summary>
        /// get the Image Format based on the MimeType
        /// </summary>
        /// <param name="mimeType">Mime type</param>
        /// <returns></returns>
        private static ImageFormat GetFormat(string mimeType)
        {
            ImageFormat result = ImageFormat.Jpeg;
            switch (mimeType)
            {
                case "image/gif":
                    result = ImageFormat.Gif;
                    break;
                case "image/png":
                    result = ImageFormat.Png;
                    break;
                case "image/bmp":
                    result = ImageFormat.Bmp;
                    break;
                case "image/tiff":
                    result = ImageFormat.Tiff;
                    break;
                case "image/x-icon":
                    result = ImageFormat.Icon;
                    break;
                default:
                    result = ImageFormat.Jpeg;
                    break;
            }
            return result;
        }
        /// <summary>
        /// get the MimeType based on the image format
        /// </summary>
        /// <param name="imageFormat">Image format</param>
        /// <returns></returns>
        private static string GetMimeType(ImageFormat imageFormat)
        {
            if (imageFormat == ImageFormat.Gif)
                return "image/gif";
            else if (imageFormat == ImageFormat.Png)
                return "image/png";
            else if (imageFormat == ImageFormat.Bmp)
                return "image/bmp";
            else if (imageFormat == ImageFormat.Tiff)
                return "image/tiff";
            else if (imageFormat == ImageFormat.Icon)
                return "image/x-icon";
            else if (imageFormat == ImageFormat.Jpeg)
                return "image/Jpeg";
            else
                return string.Empty;
        }
        #endregion

        #region Data members
        protected static readonly int THUMBNAILSIZE = 150;
        protected static readonly ImageFormat IMAGEFORMAT = ImageFormat.Jpeg;

        protected static readonly int _fontSize = 10;
        protected static readonly Color _fontColor = Color.White;
        protected static readonly Color _backgroundColor = Color.Black;
        #endregion
    }
    #endregion

    #region Document
    public partial class Document
    {
        internal static Document CreateNew(PPDataContext db, string filePath, string name, string description, string docTypeName)
        {
            // check unicity
            if (db.Documents.Count(d => d.Name.ToLower() == name.ToLower() && d.Description.ToLower() == description.ToLower()) > 0)
                throw new ExceptionDAL(string.Format("Document with Name = {0} and Description = {1} already exists", name, description));
            // check that document type exist
            DocumentType docType = db.DocumentTypes.Single(dt => dt.Name.ToLower() == docTypeName.ToLower());
            if (null == docType)
                throw new ExceptionDAL(string.Format("There is no document type with name {0}", docTypeName));
            // create new file
            File file = File.CreateNew(db, filePath);
            // build document
            Document doc = new Document();
            doc.FileID = file.ID;
            doc.Name = name;
            doc.Description = description;
            doc.DocumentType = docType;
            // insert document
            db.Documents.InsertOnSubmit(doc);
            db.SubmitChanges();
            return db.Documents.Single(d => d.Name == name && d.Description == description);
        }

        public static Document GetById(PPDataContext db, int id)
        {
            return db.Documents.SingleOrDefault(d => d.ID == id);
        }

        public Component GetComponent()
        {
            return Components.SingleOrDefault(comp => comp.DocumentID == this.ID);
        }

        public void Delete(PPDataContext db)
        {
            // delete TreeNodeDocument
            var tnDocs = db.TreeNodeDocuments.Where(tnd => tnd.DocumentID == this.ID);
            foreach (TreeNodeDocument tnDoc in tnDocs)
            {
                tnDoc.TreeNode.Delete(db, false);
                db.TreeNodeDocuments.DeleteOnSubmit(tnDoc);
                db.SubmitChanges();
            }
            
            // delete Component part
            Component component = db.Components.SingleOrDefault(cp => cp.DocumentID == this.ID);
            if (null != component)
                component.Delete(db);
            // delete Document
            Document doc = db.Documents.SingleOrDefault(d => d.ID == this.ID);
            db.Documents.DeleteOnSubmit(doc);
            db.SubmitChanges();
            // delete File
            Pic.DAL.SqlServer.File file = db.Files.SingleOrDefault(f => f.ID == this.File.ID);
            db.Files.DeleteOnSubmit(file);
            System.IO.File.Delete(file.Path);
            db.SubmitChanges();
        }

        public static void DeleteAll(PPDataContext db)
        {
            while (db.Documents.Count() > 0)
                db.Documents.First().Delete(db); 
        }
    }
    #endregion

    #region TreeNode
    public partial class TreeNode
    {
        public static TreeNode CreateNew(PPDataContext db, string name, string description, string imageFilePath)
        {
            // check file path
            if (string.IsNullOrEmpty(imageFilePath))
                imageFilePath = global::Pic.DAL.SqlServer.Properties.Settings.Default.DefaultImagePath;
            // root node already exist
            if (db.TreeNodes.Count(tn => (null == tn.ParentNodeID) && (name == tn.Name)) > 0)
                throw new ExceptionDAL(string.Format("A root node with Name = {0} already exists", name));
            // create thumbnail
            if (null == imageFilePath || imageFilePath == string.Empty)
                imageFilePath = global::Pic.DAL.SqlServer.Properties.Settings.Default.DefaultImagePath;
            Thumbnail thumbnail = Thumbnail.CreateNew(db, imageFilePath);
            // build new treenode
            TreeNode treeNode = new TreeNode();
            treeNode.Name = name;
            treeNode.Description = description;
            treeNode.Thumbnail = thumbnail;
            // insert TreeNode
            db.TreeNodes.InsertOnSubmit(treeNode);
            db.SubmitChanges();
            // retrieve new ly created thumbnail
            return treeNode;
        }

        public TreeNode CreateChild(PPDataContext db, string name, string description, string imageFilePath)
        {
            // check file path
            if (string.IsNullOrEmpty(imageFilePath))
                imageFilePath = global::Pic.DAL.SqlServer.Properties.Settings.Default.DefaultImagePath;
            // parent node already has child node with this name ?
            if (TreeNodes.Count(tn => tn.Name.ToLower() == name.ToLower()) > 0)
                throw new ExceptionDAL(string.Format("Node {0} already has child node with Name = {1}", this.Name, name));
            // create thumbnail
            if (null == imageFilePath || imageFilePath == string.Empty)
                imageFilePath = global::Pic.DAL.SqlServer.Properties.Settings.Default.DefaultImagePath;
            Thumbnail thumbnail = Thumbnail.CreateNew(db, imageFilePath);
            // build new treenode
            TreeNode treeNode = new TreeNode();
            treeNode.Name = name;
            treeNode.Description = description;
            treeNode.Thumbnail = thumbnail;
            treeNode.ParentNodeID = _ID;
            // insert TreeNode
            db.TreeNodes.InsertOnSubmit(treeNode);
            db.SubmitChanges();
            // retrieve new ly created thumbnail
            return treeNode;        
        }

        public void Move(PPDataContext db, TreeNode newParent)
        {
            if (null != newParent)
                ParentNodeID = newParent.ID;
            else
                ParentNodeID = null;
            db.SubmitChanges();
        }

        public static TreeNode GetById(PPDataContext db, int id)
        {
            TreeNode treeNode = db.TreeNodes.SingleOrDefault(tn => tn.ID == id);
            if (null == treeNode) throw new ExceptionDAL(string.Format("Failed to load TreeNode with ID = {0}", id));
            return treeNode;
        }

        public static List<TreeNode> GetRootNodes(PPDataContext db)
        {
            // retrieve rootnodes
            var nodes = db.TreeNodes.Where(tn => null == tn.ParentNodeID);
            // save in List
            List<TreeNode> listTreeNodes = new List<TreeNode>();
            foreach (TreeNode treeNode in nodes)
                listTreeNodes.Add(treeNode);
            return listTreeNodes;
        }

        public bool IsDocument
        {
            get { return (TreeNodeDocuments.Count > 0); }
        }

        public List<TreeNode> Childrens(PPDataContext db)
        {
            // get list of child nodes
            var childNodes = db.TreeNodes.Where(tn => tn.ParentNodeID == this.ID);
            // build list of child nodes
            List<TreeNode> childrens = new List<TreeNode>();
            foreach (TreeNode treeNode in childNodes)
                childrens.Add(treeNode);
            return childrens;
        }

        public DataSet ChildrenDataSet(PPDataContext db)
        {
            // Initialisation
            string connectionString = global::Pic.DAL.SqlServer.Properties.Settings.Default.PicParamDbConnectionString;
            // Cration du DataSet
            DataSet oDataSet = new DataSet("TreeNode"); // Cration et alimentation
            oDataSet.Locale = System.Globalization.CultureInfo.InvariantCulture;
            SqlDataAdapter oDaTreeNode = new SqlDataAdapter(string.Format("SELECT * FROM TreeNode WHERE ParentNodeID = {0}", this.ID), connectionString);
            oDaTreeNode.Fill(oDataSet, "TreeNode");
            return oDataSet;
         }

        public List<Document> Documents(PPDataContext db)
        {
            // get list of documents
            List<Document> documents = new List<Document>();
            var treeNodeDocuments = TreeNodeDocuments;
            foreach (TreeNodeDocument treeNodeDoc in TreeNodeDocuments)
                documents.Add(treeNodeDoc.Document);
            return documents;
        }

        public void InsertDocument(PPDataContext db, string filePath, string name, string description, string docTypeName, string thumbnailPath)
        {
            // check if this TreeNode already has documents with this same name
            if (db.TreeNodeDocuments.Count(tnd => tnd.Document.Name.ToLower() == name) > 0)
                throw new ExceptionDAL(string.Format("TreeNode {0} already has a document with name = {1}", this.Name, name));
            // get list of child nodes
            Document doc = Document.CreateNew(db, filePath, name, description, docTypeName);
            // create new TreeNode
            TreeNode tn = CreateChild(db, name, description, thumbnailPath);
            // insertion of document under treenode
            TreeNodeDocument treeNodeDoc = new TreeNodeDocument();
            treeNodeDoc.Document = doc;
            treeNodeDoc.TreeNode = tn;
            db.TreeNodeDocuments.InsertOnSubmit(treeNodeDoc);
            db.SubmitChanges();
        }

        public Component InsertComponent(PPDataContext db, string filePath, Guid guid, string name, string description, string thumbnailPath)
        {
            // check if this TreeNode already has documents with this same name
            if (db.TreeNodeDocuments.Count(tnd => tnd.Document.Name.ToLower() == name) > 0)
                throw new ExceptionDAL(string.Format("TreeNode {0} already has a document with name = {1}", this.Name, name));
            // get list of child nodes
            Document doc = Document.CreateNew(db, filePath, name, description, "Parametric component");
            // create new TreeNode
            TreeNode tn = CreateChild(db, name, description, thumbnailPath);
            // insertion of document under treenode
            TreeNodeDocument treeNodeDoc = new TreeNodeDocument();
            treeNodeDoc.Document = doc;
            treeNodeDoc.TreeNode = tn;
            db.TreeNodeDocuments.InsertOnSubmit(treeNodeDoc);
            // create component
            Component component = new Component();
            component.DocumentID = doc.ID;
            component.Guid = guid;
            db.Components.InsertOnSubmit(component);
            db.SubmitChanges();
            return db.Components.Single(c => c.Guid == guid);
        }

        public void Delete(PPDataContext db, bool deleteDocument)
        {
            // delete childrens
            foreach (TreeNode childNode in this.Childrens(db))
                childNode.Delete(db, deleteDocument);

            // delete child documents
            if (deleteDocument)
            {
                foreach (TreeNodeDocument tnDoc in TreeNodeDocuments)
                {
                    Document doc = tnDoc.Document;
                    doc.Delete(db);
                }
            }
            else
            {
                db.TreeNodeDocuments.DeleteAllOnSubmit(this.TreeNodeDocuments);
                db.SubmitChanges();
            }
            // delete record itself
            db.TreeNodes.DeleteOnSubmit(this);
            db.SubmitChanges();

            // remove thumbnail
            Thumbnail thumbnail = this.Thumbnail;
            thumbnail.Delete(db);
        }

        public static void DeleteAll(PPDataContext db)
        {
            List<TreeNode> treeNodes = TreeNode.GetRootNodes(db);
            foreach (TreeNode treeNode in treeNodes)
                treeNode.Delete(db, true);
        }
    }
    #endregion

    #region Component
    public partial class Component
    {
        public static Component GetByGuid(PPDataContext db, Guid guid)
        {
            return db.Components.SingleOrDefault(c => c.Guid == guid);
        }

        public static Component GetByDocumentID(PPDataContext db, int documentId)
        {
            return db.Components.Single(c => c.DocumentID == documentId);
        }

        public MajorationSet GetMajorationSetByProfile(PPDataContext db, CardboardProfile profile)
        {
            return db.MajorationSets.SingleOrDefault(mjs => mjs.CardboardProfileID == profile.ID && mjs.ComponentID == this.ID);
        }

        public Dictionary<string, double> GetDefaultMajorations(CardboardProfile profile)
        {
            if (null == profile)
                throw new ExceptionDAL("Profile is null reference!");
            // find nearest set
            MajorationSet nearestSet = null;
            foreach (MajorationSet majoSet in this.MajorationSets)
            {
                if (null == nearestSet
                    || (Math.Abs(majoSet.CardboardProfile.Thickness - profile.Thickness) < Math.Abs(nearestSet.CardboardProfile.Thickness - profile.Thickness))
                    )
                    nearestSet = majoSet;
            }

            // build dictionnary
            Dictionary<string, double> dict = new Dictionary<string, double>();
            if (null != nearestSet)
            {
                double coef = (double)(profile.Thickness / nearestSet.CardboardProfile.Thickness);
                foreach (Majoration maj in nearestSet.Majorations)
                    dict.Add(maj.Name, maj.Value * coef);
            }
            return dict;
        }

        public void InsertNewMajorationSet(PPDataContext db, string profileName, Dictionary<string, double> majorations)
        {
            CardboardProfile profile = db.CardboardProfiles.Single(cbp => cbp.Name.ToLower() == profileName.ToLower());
            if (null == profile)
                throw new ExceptionDAL(string.Format("There is no profile with name = {0}", profileName));
            MajorationSet majoSetExisting = GetMajorationSetByProfile(db, profile);
            if (null != majoSetExisting)
                throw new ExceptionDAL(string.Format("A majoration set for component {0} and profile {1} already exists", this.Document.Name, profileName));
            MajorationSet majoSet = new MajorationSet();
            majoSet.ComponentID = this.ID;
            majoSet.CardboardProfile = db.CardboardProfiles.Single(cbp => cbp.Name.ToLower() == profileName.ToLower());
            db.MajorationSets.InsertOnSubmit(majoSet);
            db.SubmitChanges();

            // get majoset
            MajorationSet majoSetNew = db.MajorationSets.SingleOrDefault(mjs => mjs.CardboardProfileID == profile.ID && mjs.ComponentID == this.ID);
            foreach (KeyValuePair<string, double> pair in majorations)
            {
                Majoration maj = new Majoration();
                maj.MajorationSet = majoSetNew;
                maj.Name = pair.Key;
                maj.Value = (float)pair.Value;
                db.Majorations.InsertOnSubmit(maj);
            }
            db.SubmitChanges();
        }

        public void UpdateMajorationSet(PPDataContext db, CardboardProfile profile, Dictionary<string, double> majorations)
        {
            MajorationSet majoSet = db.MajorationSets.SingleOrDefault(mjs => mjs.CardboardProfileID == profile.ID && mjs.ComponentID == this.ID);
            foreach (Majoration maj in majoSet.Majorations)
                maj.Value = (float)majorations[maj.Name];
            db.SubmitChanges();
        }

        public void Delete(PPDataContext db)
        {
            // delete majoration sets
            while (MajorationSets.Count() > 0)
                MajorationSets.First().Delete(db);
            // delete component itself
            Component comp = db.Components.Single(cp => cp.ID == this.ID);
            db.Components.DeleteOnSubmit(comp);
            db.SubmitChanges();
            // delete document
            Document.Delete(db);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(string.Format("Component Guid = {0},  Name = {1}", this.Guid.ToString(), this.Document.Name));
            foreach (MajorationSet majSet in this.MajorationSets)
            {
                sb.AppendLine(string.Format("\tProfile {0}", majSet.CardboardProfile.Name));
                foreach (Majoration majo in majSet.Majorations)
                    sb.AppendLine(string.Format("\t\t{0} = {1}", majo.Name, majo.Value));
            }
            return sb.ToString();
        }

        public static void DeleteAll(PPDataContext db)
        {
            while (db.Components.Count() > 0)
                db.Components.First().Delete(db);
        }
    }
    #endregion
}
